# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import six
import uuid
import base64
import rfc3986
from ..config import ERROR


# Module API

def cast_string(format, value):
    if not isinstance(value, six.string_types):
        return ERROR
    if format == 'uri':
        if not rfc3986.is_valid_uri(value, require_scheme=True):
            return ERROR
    elif format == 'email':
        if not re.match(_EMAIL_PATTERN, value):
            return ERROR
    elif format == 'uuid':
        try:
            uuid.UUID(value, version=4)
        except Exception:
            return ERROR
    elif format == 'binary':
        try:
            base64.b64decode(value)
        except Exception:
            return ERROR
    return value


# Internal

_EMAIL_PATTERN = re.compile(r'[^@]+@[^@]+\.[^@]+')



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
from collections import namedtuple
from ..config import ERROR


# Module API

def cast_yearmonth(format, value):
    if isinstance(value, (tuple, list)):
        if len(value) != 2:
            return ERROR
        value = _yearmonth(value[0], value[1])
    elif isinstance(value, six.string_types):
        try:
            year, month = value.split('-')
            year = int(year)
            month = int(month)
            if month < 1 or month > 12:
                return ERROR
            value = _yearmonth(year, month)
        except Exception:
            return ERROR
    else:
        return ERROR
    return value


# Internal

_yearmonth = namedtuple('yearmonth', ['year', 'month'])



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', True, True),
    ('default', 'yes', True),
    ('default', 'y', True),
    ('default', 'true', True),
    ('default', 't', True),
    ('default', '1', True),
    ('default', 'YES', True),
    ('default', 'Yes', True),
    ('default', False, False),
    ('default', 'no', False),
    ('default', 'n', False),
    ('default', 'false', False),
    ('default', 'f', False),
    ('default', '0', False),
    ('default', 'NO', False),
    ('default', 'No', False),
    ('default', 0, ERROR),
    ('default', 1, ERROR),
    ('default', '3.14', ERROR),
    ('default', '', ERROR),
])
def test_cast_boolean(format, value, result):
    assert types.cast_boolean(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default',
        {'properties': {'Ã': 'Ã'}, 'type': 'Feature', 'geometry': None},
        {'properties': {'Ã': 'Ã'}, 'type': 'Feature', 'geometry': None}),
    ('default',
        '{"geometry": null, "type": "Feature", "properties": {"\\u00c3": "\\u00c3"}}',
        {'properties': {'Ã': 'Ã'}, 'type': 'Feature', 'geometry': None}),
    ('default', {'coordinates': [0, 0, 0], 'type': 'Point'}, ERROR),
    ('default', 'string', ERROR),
    ('default', 1, ERROR),
    ('default', '3.14', ERROR),
    ('default', '', ERROR),
    ('default', {}, ERROR),
    ('default', '{}', ERROR),
    ('topojson',
        {'type': 'LineString', 'arcs': [42]},
        {'type': 'LineString', 'arcs': [42]}),
    ('topojson',
        '{"type": "LineString", "arcs": [42]}',
        {'type': 'LineString', 'arcs': [42]}),
    ('topojson', 'string', ERROR),
    ('topojson', 1, ERROR),
    ('topojson', '3.14', ERROR),
    ('topojson', '', ERROR),
])
def test_cast_geojson(format, value, result):
    assert types.cast_geojson(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import constraints


# Tests

@pytest.mark.parametrize('constraint, value, result', [
    (0, 1, False),
    (1, 1, True),
    (2, 1, True),
])
def test_check_maximum(constraint, value, result):
    assert constraints.check_maximum(constraint, value) == result



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


# Module API

def check_minimum(constraint, value):
    if value is None:
        return True
    if value >= constraint:
        return True
    return False



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import constraints


# Tests

@pytest.mark.parametrize('constraint, value, result', [
    (False, 'any', True),
    (True, 'any', True),
])
def test_check_unique(constraint, value, result):
    assert constraints.check_unique(constraint, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from datetime import date
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', date(2019, 1, 1), date(2019, 1, 1)),
    ('default', '2019-01-01', date(2019, 1, 1)),
    ('default', '10th Jan 1969', ERROR),
    ('default', 'invalid', ERROR),
    ('default', True, ERROR),
    ('default', '', ERROR),
    ('any', date(2019, 1, 1), date(2019, 1, 1)),
    ('any', '2019-01-01', date(2019, 1, 1)),
    ('any', '10th Jan 1969', date(1969, 1, 10)),
    ('any', '10th Jan nineteen sixty nine', ERROR),
    ('any', 'invalid', ERROR),
    ('any', True, ERROR),
    ('any', '', ERROR),
    ('%d/%m/%y', date(2019, 1, 1), date(2019, 1, 1)),
    ('%d/%m/%y', '21/11/06', date(2006, 11, 21)),
    ('%y/%m/%d','21/11/06 16:30', ERROR),
    ('%d/%m/%y','invalid', ERROR),
    ('%d/%m/%y',True, ERROR),
    ('%d/%m/%y', '', ERROR),
    ('invalid','21/11/06 16:30', ERROR),
    # Deprecated
    ('fmt:%d/%m/%y', date(2019, 1, 1), date(2019, 1, 1)),
    ('fmt:%d/%m/%y', '21/11/06', date(2006, 11, 21)),
    ('fmt:%y/%m/%d','21/11/06 16:30', ERROR),
    ('fmt:%d/%m/%y','invalid', ERROR),
    ('fmt:%d/%m/%y',True, ERROR),
    ('fmt:%d/%m/%y', '', ERROR),
])
def test_cast_date(format, value, result):
    assert types.cast_date(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import constraints


# Tests

@pytest.mark.parametrize('constraint, value, result', [
    (0, [1], False),
    (1, [1], True),
    (2, [1], True),
])
def test_check_maxLength(constraint, value, result):
    assert constraints.check_maxLength(constraint, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import os
import json


# Internal

def _load_spec(filename):
    path = os.path.join(os.path.dirname(__file__), 'specs', filename)
    spec = json.load(io.open(path, encoding='utf-8'))
    return spec


# Public API

table_schema = _load_spec('table-schema.json')
geojson = _load_spec('geojson.json')



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', 'string', 'string'),
    ('default', '', ''),
    ('default', 0, ERROR),
    ('uri', 'http://google.com', 'http://google.com'),
    ('uri', 'string', ERROR),
    ('uri', '', ERROR),
    ('uri', 0, ERROR),
    ('email', 'name@gmail.com', 'name@gmail.com'),
    ('email', 'http://google.com', ERROR),
    ('email', 'string', ERROR),
    ('email', '', ERROR),
    ('email', 0, ERROR),
    ('binary', 'dGVzdA==', 'dGVzdA=='),
    ('binary', '', ''),
    ('binary', 'string', ERROR),
    ('binary', 0, ERROR),
])
def test_cast_string(format, value, result):
    assert types.cast_string(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import deepcopy
from functools import partial
from . import constraints
from . import exceptions
from . import config
from . import specs
from . import types


# Module API

class Field(object):
    """Table Schema field representation.
    """

    # Public

    def __init__(self, descriptor, missing_values=config.DEFAULT_MISSING_VALUES):

        # Deepcopy descriptor
        descriptor = deepcopy(descriptor)

        # Apply descriptor defaults
        descriptor.setdefault('type', config.DEFAULT_FIELD_TYPE)
        descriptor.setdefault('format', config.DEFAULT_FIELD_FORMAT)

        # Set attributes
        self.__descriptor = descriptor
        self.__missing_values = missing_values
        self.__cast_function = self.__get_cast_function()
        self.__check_functions = self.__get_check_functions()

    @property
    def descriptor(self):
        """dict: field descriptor
        """
        return self.__descriptor

    @property
    def name(self):
        """str: field name
        """
        return self.__descriptor['name']

    @property
    def type(self):
        """str: field type
        """
        return self.__descriptor['type']

    @property
    def format(self):
        """str: field format
        """
        return self.__descriptor['format']

    @property
    def constraints(self):
        """dict: field constraints
        """
        return self.__descriptor.get('constraints', {})

    @property
    def required(self):
        """bool: true if field is required
        """
        return self.constraints.get('required', False)

    def cast_value(self, value, constraints=True):
        """Cast value against field.

        Args:
            value (any): value to cast
            constraints (None/str[]/False):
                - pass True to check all constraints (default)
                - pass list of constraints for granular check
                - pass False to skip all constraints

        Returns:
            any: cast value

        """

        # Null value
        if value in self.__missing_values:
            value = None

        # Cast value
        cast_value = value
        if value is not None:
            cast_value = self.__cast_function(value)
            if cast_value == config.ERROR:
                raise exceptions.InvalidCastError((
                    'Field "{field.name}" can\'t cast value "{value}" '
                    'for type "{field.type}" with format "{field.format}"'
                    ).format(field=self, value=value))

        # Check value
        if constraints:
            for name, check in self.__check_functions.items():
                if isinstance(constraints, list):
                    if name not in constraints:
                        continue
                passed = check(cast_value)
                if not passed:
                    raise exceptions.ConstraintError((
                        'Field "{field.name}" has constraint "{name}" '
                        'which is not satisfied for value "{value}"'
                        ).format(field=self, name=name, value=value))

        return cast_value

    def test_value(self, value, constraints=True):
        """Cast value against field.

        Args:
            value (mixed): value to test
            constraints (None/str[]/False):
                - pass True to check all constraints (default)
                - pass list of constraints for granular check
                - pass False to skip all constraints

        Returns:
            bool: result of test

        """
        try:
            self.cast_value(value, constraints=constraints)
        except (exceptions.InvalidCastError, exceptions.ConstraintError):
            return False
        return True

    # Private

    def __get_cast_function(self):
        options = {}
        # Get cast options for number
        if self.type == 'number':
            for key in ['decimalChar', 'groupChar', 'currency']:
                value = self.descriptor.get(key)
                if value is not None:
                    options[key] = value
        cast = getattr(types, 'cast_%s' % self.type)
        cast = partial(cast, self.format, **options)
        return cast

    def __get_check_functions(self):
        checks = {}
        cast = partial(self.cast_value, constraints=False)
        whitelist = _get_field_constraints(specs.table_schema, self.type)
        for name, constraint in self.constraints.items():
            if name in whitelist:
                # Cast enum constraint
                if name in ['enum']:
                    constraint = list(map(cast, constraint))
                # Cast maximum/minimum constraint
                if name in ['maximum', 'minimum']:
                    constraint = cast(constraint)
                check = getattr(constraints, 'check_%s' % name)
                checks[name] = partial(check, constraint)
        return checks


# Internal

def _get_field_constraints(spec, type):
    # Extract list of constraints for given type from jsonschema
    spec_types = spec['properties']['fields']['items']['anyOf']
    for spec_type in spec_types:
        if type in spec_type['properties']['type']['enum']:
            return spec_type['properties']['constraints']['properties'].keys()



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from tabulator import Stream
from functools import partial
from importlib import import_module
from .schema import Schema
from .infer import infer
from . import compat


# Module API

class Table(object):
    """Table Schema table representation.

    Args:
        source (mixed): data source
        schema (Schema/dict/str): schema instance or descriptor/path/url
        backend (None/str): backend name like `sql` or `bigquery`
        options (dict): tabulator options or backend options

    """

    # Public

    def __init__(self, source, schema=None, post_cast=None,
                 backend=None, **options):

        # Defaults
        if post_cast is None:
            post_cast = []

        # Schema
        self.__schema = None
        if isinstance(schema, (compat.str, dict)):
            self.__schema = Schema(schema)

        # Tabulator
        if backend is None:
            options.setdefault('headers', 1)
            self.__stream = Stream(source,  **options)
            self.__stream.open()
            if self.__schema is None:
                self.__schema = Schema(infer(
                    self.__stream.headers, self.__stream.sample))

        # Storage
        else:
            module = 'tableschema.plugins.%s' % backend
            storage = import_module(module).Storage(**options)
            generator = partial(storage.iter, source)
            if self.__schema is None:
                self.__schema = Schema(storage.describe(source))
            storage.describe(source, self.__schema.descriptor)
            self.__stream = Stream(generator, headers=self.__schema.headers)
            self.__stream.open()

        # Attributes
        self.__post_cast = post_cast

    @property
    def stream(self):
        """tabulator.Stream: stream instance
        """
        return self.__stream

    @property
    def schema(self):
        """Schema: schema instance
        """
        return self.__schema

    def iter(self, keyed=False, extended=False):
        """Yields table rows.

        Args:
            keyed (bool): yield keyed rows
            extended (bool): yield extended rows

        Yields:
            mixed[]/mixed{}: row or keyed row or extended row

        """
        self.__stream.reset()
        iterator = self.__stream.iter(extended=True)
        iterator = self.__apply_processors(iterator)
        for number, headers, row in iterator:
            if extended:
                yield (number, headers, row)
            elif keyed:
                yield dict(zip(headers, row))
            else:
                yield row

    def read(self, keyed=False, extended=False, limit=None):
        """Read table rows.

        Args:
            limit (int): return this amount of rows

        Returns:
            list[]: table rows

        """
        result = []
        rows = self.iter(keyed=keyed, extended=extended)
        for count, row in enumerate(rows, start=1):
            result.append(row)
            if count == limit:
                break
        return result

    def save(self, target, backend=None, **options):
        """Save table rows.

        NOTE: To save schema use `table.schema.save(target)`

        Args:
            target (str): saving target
            backend (None/str): backend name like sql` or `bigquery`
            options (dict): tabulator options or backend options

        Returns:
            None/Storage: storage instance if backend used

        """

        # Tabulator
        if backend is None:
            with Stream(self.iter, headers=self.__schema.headers) as stream:
                stream.save(target, **options)

        # Storage
        else:
            module = 'tableschema.plugins.%s' % backend
            storage = import_module(module).Storage(**options)
            storage.create(target, self.__schema.descriptor, force=True)
            storage.write(target, self.iter())
            return storage

    # Internal

    def __apply_processors(self, iterator):

        # Apply processors to iterator
        def builtin_processor(extended_rows):
            for number, headers, row in extended_rows:
                headers = self.__schema.headers
                row = self.__schema.cast_row(row)
                yield (number, headers, row)
        processors = [builtin_processor] + self.__post_cast
        for processor in processors:
            iterator = processor(iterator)
        return iterator



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', 1, 1),
    ('default', '1', '1'),
    ('default', '3.14', '3.14'),
    ('default', True, True),
    ('default', '', ''),
])
def test_cast_any(format, value, result):
    assert types.cast_any(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from decimal import Decimal
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, options, result', [
    ('default', Decimal(1), {}, Decimal(1)),
    ('default', 1, {}, Decimal(1)),
    ('default', 1.0, {}, Decimal(1)),
    ('default', '1', {}, Decimal(1)),
    ('default', '10.00', {}, Decimal(10)),
    ('default', '10.50', {}, Decimal(10.5)),
    ('default', '100%', {}, Decimal(1)),
    ('default', '1000‰', {}, Decimal(10)),
    ('default', '-1000', {}, Decimal(-1000)),
    ('default', '1,000', {'groupChar': ','}, Decimal(1000)),
    ('default', '10,000.00', {'groupChar': ','}, Decimal(10000)),
    ('default', '10,000,000.50', {'groupChar': ','}, Decimal(10000000.5)),
    ('default', '10#000.00', {'groupChar': '#'}, Decimal(10000)),
    ('default', '10#000#000.50', {'groupChar': '#'}, Decimal(10000000.5)),
    ('default', '10.50', {'groupChar': '#'}, Decimal(10.5)),
    ('default', '1#000', {'groupChar': '#'}, Decimal(1000)),
    ('default', '10#000@00', {'groupChar': '#', 'decimalChar': '@'}, Decimal(10000)),
    ('default', '10#000#000@50', {'groupChar': '#', 'decimalChar': '@'}, Decimal(10000000.5)),
    ('default', '10@50', {'groupChar': '#', 'decimalChar': '@'}, Decimal(10.5)),
    ('default', '1#000', {'groupChar': '#', 'decimalChar': '@'}, Decimal(1000)),
    ('default', '10,000.00', {'groupChar': ',', 'currency': True}, Decimal(10000)),
    ('default', '10,000,000.00', {'groupChar': ',', 'currency': True}, Decimal(10000000)),
    ('default', '$10000.00', {'currency': True}, Decimal(10000)),
    ('default', '  10,000.00 €', {'groupChar': ',', 'currency': True}, Decimal(10000)),
    ('default', '10 000,00', {'groupChar': ' ', 'decimalChar': ','}, Decimal(10000)),
    ('default', '10 000 000,00', {'groupChar': ' ', 'decimalChar': ','}, Decimal(10000000)),
    ('default', '10000,00 ₪', {'groupChar': ' ', 'decimalChar': ',', 'currency': True}, Decimal(10000)),
    ('default', '  10 000,00 £', {'groupChar': ' ', 'decimalChar': ',', 'currency': True}, Decimal(10000)),
    ('default', '10,000a.00', {}, ERROR),
    ('default', '10+000.00', {}, ERROR),
    ('default', '$10:000.00', {}, ERROR),
    ('default', 'string', {}, ERROR),
    ('default', '', {}, ERROR),
])
def test_cast_number(format, value, options, result):
    assert types.cast_number(format, value, **options) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', (180, 90), (180, 90)),
    ('default', [180, 90], (180, 90)),
    ('default', '180,90', (180, 90)),
    ('default', '180, -90', (180, -90)),
    ('default', {'lon': 180, 'lat': 90}, ERROR),
    ('default', '181,90', ERROR),
    ('default', '0,91', ERROR),
    ('default', 'string', ERROR),
    ('default', 1, ERROR),
    ('default', '3.14', ERROR),
    ('default', '', ERROR),
    ('array', (180, 90), (180, 90)),
    ('array', [180, 90], (180, 90)),
    ('array', '[180, -90]', (180, -90)),
    ('array', {'lon': 180, 'lat': 90}, ERROR),
    ('array', [181, 90], ERROR),
    ('array', [0, 91], ERROR),
    ('array', '180,90', ERROR),
    ('array', 'string', ERROR),
    ('array', 1, ERROR),
    ('array', '3.14', ERROR),
    ('array', '', ERROR),
    ('object', {'lon': 180, 'lat': 90}, (180, 90)),
    ('object', '{"lon": 180, "lat": 90}', (180, 90)),
    ('object', '[180, -90]', ERROR),
    ('object', {'lon': 181, 'lat': 90}, ERROR),
    ('object', {'lon': 180, 'lat': -91}, ERROR),
    ('object', [180, -90], ERROR),
    ('object', '180,90', ERROR),
    ('object', 'string', ERROR),
    ('object', 1, ERROR),
    ('object', '3.14', ERROR),
    ('object', '', ERROR),
])
def test_cast_geopoint(format, value, result):
    assert types.cast_geopoint(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import jsonschema
from jsonschema.validators import validator_for
from . import exceptions
from . import specs


# Module API


def validate(descriptor, no_fail_fast=False):
    """Validate Table Schema schema descriptor.

    Args:
        descriptor (dict): schema descriptor to validate
        no_fail_fast (bool): collect all errors

    Raises:
        exceptions.SchemaValidationError
        exceptions.MultipleInvalid (no_fail_fast=True)

    Returns:
        bool: True

    """

    # Fail fast
    if not no_fail_fast:
        jsonschema.validate(
            descriptor, specs.table_schema, cls=_TableSchemaValidator)

    # Multiple errors
    else:
        validator = _TableSchemaValidator(specs.table_schema)
        errors = list(validator.iter_errors(descriptor))
        if errors:
            raise exceptions.MultipleInvalid(errors=errors)

    return True


# Internal


class _TableSchemaValidator(validator_for(specs.table_schema)):
    @classmethod
    def check_schema(cls, schema):
        # When checking against the metaschema, we do not want to run the
        # additional checking added in iter_errors
        parent_cls = cls.__bases__[0]
        for error in parent_cls(cls.META_SCHEMA).iter_errors(schema):
            raise jsonschema.exceptions.SchemaError.create_from(error)

    def iter_errors(self, instance, _schema=None):

        for e in super(_TableSchemaValidator, self).iter_errors(
                instance, _schema):
            yield exceptions.SchemaValidationError(
                e.message, e.validator, e.path, e.cause, e.context,
                e.validator_value, e.instance, e.schema, e.schema_path,
                e.parent)

        try:
            field_names = [f['name'] for f in instance['fields']]
        except (TypeError, KeyError):
            field_names = []

        # the hash MAY contain a key `primaryKey`
        if isinstance(instance, dict) and instance.get('primaryKey'):

            # ensure that the primary key matches field names
            if isinstance(instance['primaryKey'], list):
                for k in instance['primaryKey']:
                    if k not in field_names:
                        yield exceptions.SchemaValidationError(
                            'A Table Schema primaryKey value must be '
                            'found in the schema field names')

        # the hash may contain a key `foreignKeys`
        if isinstance(instance, dict) and instance.get('foreignKeys'):
            for fk in instance['foreignKeys']:

                # ensure that `foreignKey.fields` match field names
                if isinstance(fk.get('fields', []), list):
                    for field in fk.get('fields'):
                        if field not in field_names:
                            yield exceptions.SchemaValidationError(
                                'A Table Schema foreignKey.fields value '
                                'must correspond with field names.')
                    if not (len(fk.get('fields')) ==
                            len(fk['reference']['fields'])):
                        yield exceptions.SchemaValidationError(
                            'A Table Schema foreignKey.fields must '
                            'contain the same number entries as '
                            'foreignKey.reference.fields.')



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from datetime import datetime
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', datetime(2014, 1, 1, 6), datetime(2014, 1, 1, 6)),
    ('default', '2014-01-01T06:00:00Z', datetime(2014, 1, 1, 6)),
    ('default', 'Mon 1st Jan 2014 9 am', ERROR),
    ('default', 'invalid', ERROR),
    ('default', True, ERROR),
    ('default', '', ERROR),
    ('any', datetime(2014, 1, 1, 6), datetime(2014, 1, 1, 6)),
    ('any', '10th Jan 1969 9 am', datetime(1969, 1, 10, 9)),
    ('any', 'invalid', ERROR),
    ('any', True, ERROR),
    ('any', '', ERROR),
    ('%d/%m/%y %H:%M', datetime(2006, 11, 21, 16, 30), datetime(2006, 11, 21, 16, 30)),
    ('%d/%m/%y %H:%M', '21/11/06 16:30', datetime(2006, 11, 21, 16, 30)),
    ('%H:%M %d/%m/%y', '21/11/06 16:30', ERROR),
    ('%d/%m/%y %H:%M', 'invalid', ERROR),
    ('%d/%m/%y %H:%M', True, ERROR),
    ('%d/%m/%y %H:%M', '', ERROR),
    ('invalid', '21/11/06 16:30', ERROR),
    # Deprecated
    ('fmt:%d/%m/%y %H:%M', datetime(2006, 11, 21, 16, 30), datetime(2006, 11, 21, 16, 30)),
    ('fmt:%d/%m/%y %H:%M', '21/11/06 16:30', datetime(2006, 11, 21, 16, 30)),
    ('fmt:%H:%M %d/%m/%y', '21/11/06 16:30', ERROR),
    ('fmt:%d/%m/%y %H:%M', 'invalid', ERROR),
    ('fmt:%d/%m/%y %H:%M', True, ERROR),
    ('fmt:%d/%m/%y %H:%M', '', ERROR),
])
def test_cast_datetime(format, value, result):
    assert types.cast_datetime(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', [2000, 10], (2000, 10)),
    ('default', (2000, 10), (2000, 10)),
    ('default', '2000-10', (2000, 10)),
    ('default', (2000, 10, 20), ERROR),
    ('default', '2000-13-20', ERROR),
    ('default', '2000-13', ERROR),
    ('default', '2000-0', ERROR),
    ('default', '13', ERROR),
    ('default', -10, ERROR),
    ('default', 20, ERROR),
    ('default', '3.14', ERROR),
    ('default', '', ERROR),
])
def test_cast_yearmonth(format, value, result):
    assert types.cast_yearmonth(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
from setuptools import setup, find_packages


# Helpers
def read(*paths):
    """Read a text file."""
    basedir = os.path.dirname(__file__)
    fullpath = os.path.join(basedir, *paths)
    contents = io.open(fullpath, encoding='utf-8').read().strip()
    return contents


# Prepare
PACKAGE = 'tableschema'
NAME = PACKAGE.replace('_', '-')
INSTALL_REQUIRES = [
    'six>=1.9,<2.0a',
    'click>=3.3,<7.0a',
    'requests>=2.5,<3.0a',
    'python-dateutil>=2.4,<3.0a',
    'jsonschema>=2.5,<3.0a',
    'rfc3986>=0.4,<1.0a',
    'future>=0.15,<1.0a',
    'unicodecsv>=0.14,<1.0a',
    'tabulator>=0.7,<1.0a',
    'isodate>=0.5.4,<1.0a',
]
TESTS_REQUIRE = [
    'pylama',
    'tox',
]
README = read('README.md')
VERSION = read(PACKAGE, 'VERSION')
PACKAGES = find_packages(exclude=['examples', 'tests'])


# Run
setup(
    name=NAME,
    version=VERSION,
    packages=PACKAGES,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require={'develop': TESTS_REQUIRE},
    entry_points={
        'console_scripts': [
            'tableschema = tableschema.cli:main',
        ]
    },
    zip_safe=False,
    long_description=README,
    description='A utility library for working with Table Schema in Python',
    author='Open Knowledge Foundation',
    author_email='info@okfn.org',
    url='https://github.com/frictionlessdata/tableschema-py',
    license='MIT',
    keywords=[
        'frictionless data',
        'open data',
        'json schema',
        'table schema',
        'data package',
        'tabular data package',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
import json
import jsonschema
from .. import specs
from ..config import ERROR


# Module API

def cast_geojson(format, value):
    if not isinstance(value, dict):
        if not isinstance(value, six.string_types):
            return ERROR
        try:
            value = json.loads(value)
        except Exception:
            return ERROR
    if format == 'default':
        try:
            jsonschema.validate(value, specs.geojson)
        except Exception:
            return ERROR
    elif format == 'topojson':
        if not isinstance(value, dict):
            return ERROR
    return value



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import pytest
from tableschema import types
from tableschema.config import ERROR


# Tests

@pytest.mark.parametrize('format, value, result', [
    ('default', 2000, 2000),
    ('default', '2000', 2000),
    ('default', -2000, ERROR),
    ('default', 20000, ERROR),
    ('default', '3.14', ERROR),
    ('default', '', ERROR),
])
def test_cast_year(format, value, result):
    assert types.cast_year(format, value) == result



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import six
import unicodedata
from decimal import Decimal
from ..config import ERROR
from .. import compat


# Module API

def cast_number(format, value, **options):
    percentage = False
    currency = options.get('currency', False)
    group_char = options.get('groupChar', _DEFAULT_GROUP_CHAR)
    decimal_char = options.get('decimalChar', _DEFAULT_DECIMAL_CHAR)
    if not isinstance(value, Decimal):
        if isinstance(value, six.string_types):
            value = re.sub('\s', '', value)
            value = value.replace(decimal_char, '.')
            value = value.replace(group_char, '')
            result = re.sub('[' + _PERCENT_CHAR + ']', '', value)
            if value != result:
                value = result
                percentage = True
            if currency:
                pattern = '[{0}]'.format(_CURRENCIES)
                value = re.sub(pattern, '', value)
        elif not isinstance(value, (int, float)):
            return ERROR
        try:
            value = Decimal(value)
        except Exception:
            return ERROR
    if percentage:
        value = value/100
    return value


# Internal

_DEFAULT_DECIMAL_CHAR = '.'
_DEFAULT_GROUP_CHAR = ''
_PERCENT_CHAR = '%‰‱％﹪٪'
_CURRENCIES = ''.join(compat.chr(i) for i in range(0xffff)
     if unicodedata.category(compat.chr(i)) == 'Sc')
